/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.Console;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.Objects;

/**
 * An <i>adapter</i> that adapts a {@link Console} to a {@link ProgramConsole}.
 * <p>
 * This class is conditionally thread-safe. Since it wraps a {@code Console},
 * it has the same possibility for some calls to block in multithreaded
 * scenarios as noted in the {@code Console} class documentation.
 */
public final class ConsoleProgramConsole implements ProgramConsole {
  private static final ConsoleProgramConsole SYSTEM_INSTANCE = computeSystemInstance();

  private final Console adaptee;

  /**
   * Constructs an instance with the specified console.
   *
   * @param adaptee the {@code Console} to adapt to this interface
   *
   * @throws NullPointerException if {@code adaptee} is {@code null}
   */
  public ConsoleProgramConsole(Console adaptee) {
    super();
    this.adaptee = Objects.requireNonNull(adaptee, "adaptee is null");
  }

  @Override
  public PrintWriter writer() {
    return adaptee.writer();
  }

  @Override
  public Reader reader() {
    return adaptee.reader();
  }

  @Override
  public ConsoleProgramConsole format(String fmt, Object... args) {
    adaptee.format(fmt, args);
    return this;
  }

  @Override
  public ConsoleProgramConsole printf(String format, Object... args) {
    adaptee.printf(format, args);
    return this;
  }

  @Override
  public String readLine(String fmt, Object... args) {
    return adaptee.readLine(fmt, args);
  }

  @Override
  public String readLine() {
    return adaptee.readLine();
  }

  @Override
  public char[] readPassword(String fmt, Object... args) {
    return adaptee.readPassword(fmt, args);
  }

  @Override
  public char[] readPassword() {
    return adaptee.readPassword();
  }

  @Override
  public void flush() {
    adaptee.flush();
  }

  /**
   * Returns the system {@code ConsoleProgramConsole} instance for the system
   * {@code Console}.
   *
   * @return the system instance, or {@code null} if there is no system
   *         {@code Console}
   */
  public static ConsoleProgramConsole systemInstance() {
    return SYSTEM_INSTANCE;
  }

  private static ConsoleProgramConsole computeSystemInstance() {
    return (System.console() == null) ? null : new ConsoleProgramConsole(System.console());
  }
}
