/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.Flushable;

/**
 * A computer program.
 */
public interface Program extends Flushable {
  /**
   * Exit status {@code 0} indicating success.
   */
  static final int EXIT_SUCCESS = 0;
  /**
   * Exit status {@code 1} indicating failure.
   */
  static final int EXIT_FAILURE = 1;

  /**
   * Runs this program.
   *
   * @return the exit status of this program
   *
   * @throws Exception if an exceptional condition occurs
   */
  int run() throws Exception;

  /**
   * Prints an error for the specified exception.
   * <p>
   * The error is considered to be printed fully (e.g., flushed) when the call
   * returns.
   * <p>
   * Clients may ignore any exception thrown by this method.
   *
   * @param t the exception for which an error should be printed
   *
   * @throws StandardIoException if an error occurs while writing to an
   *           underlying standard I/O destination
   */
  void printError(Throwable t);

  /**
   * Returns an exit status for the specified exception.
   * <p>
   * Clients may ignore any exception thrown by this method.
   *
   * @param t the exception for which an exit status should be returned
   *
   * @return the exit status
   */
  int exitStatus(Throwable t);

  /**
   * Flushes any buffered output to underlying standard I/O destinations.
   *
   * @throws StandardIoException if an error occurs while flushing an
   *           underlying standard I/O destination
   */
  @Override
  void flush();
}
