/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.InterruptedIOException;
import java.io.PrintWriter;
import java.nio.channels.ClosedByInterruptException;
import java.util.EnumSet;
import java.util.Objects;
import java.util.Set;

import org.imca_cat.prgm.internal.NullWriter;

/**
 * A skeletal computer program that behaves in a sensible way.
 * <p>
 * Sensible behavior is achieved via a {@link SensibleUi}.
 * <p>
 * The only method that must be overridden in a subclass is
 * {@link Program#run() run()}; this is the entry point of the program.
 * <p>
 * This program synchronizes critical sections on the object returned by
 * {@link #lock() lock()}; subclasses that need to synchronize their own
 * critical sections on the same object can obtain it via this method.
 * <p>
 * Here's an example "Hello World" program:
 *
 * <pre>
 * public class HelloWorld {
 *   public static void main(String[] args) {
 *     new SensibleProgram&lt;SensibleUi&gt;(DefaultConfigurableSensibleUi.create()) {
 *       &#064;Override
 *       public int run() throws Exception {
 *         ui().write("%s%n", "hello, world");
 *         return 0;
 *       }
 *     }.runAndExitVm();
 *   }
 * }</pre>
 * <p>
 * This class is thread-safe.
 *
 * @param <T> the type of this program's UI
 */
public abstract class SensibleProgram<T extends SensibleUi> extends AbstractProgram {
  private static final PrintWriter NULL_PRINT_WRITER = new PrintWriter(new NullWriter());
  private static final Set<AbstractProgram.Option> EMPTY_OPTIONS = EnumSet
      .noneOf(AbstractProgram.Option.class);

  private T ui;

  /**
   * Constructs an instance with an {@link Object} critical section lock and
   * the specified UI.
   *
   * @param ui the user interface of this program; this program assumes
   *          ownership of the UI; the client should not maintain references to
   *          it and must not use or modify it in any way
   *
   * @throws NullPointerException if {@code ui} is {@code null}
   */
  public SensibleProgram(T ui) {
    this(new Object(), ui);
  }

  /**
   * Constructs an instance with the specified critical section lock and UI.
   *
   * @param lock the object on which this program's critical sections
   *          synchronize
   * @param ui the user interface of this program; this program assumes
   *          ownership of the UI; the client should not maintain references to
   *          it and must not use or modify it in any way
   *
   * @throws NullPointerException if {@code lock} or {@code ui} is {@code null}
   */
  public SensibleProgram(Object lock, T ui) {
    super(lock, NULL_PRINT_WRITER, EMPTY_OPTIONS);
    this.ui = Objects.requireNonNull(ui, "ui is null");
  }

  /**
   * {@inheritDoc}
   *
   * @implSpec This implementation prints nothing if the exception is an
   *           {@link InterruptedException}, {@link InterruptedIOException}, or
   *           {@link ClosedByInterruptException}. Otherwise, it calls
   *           {@link #lock() lock()} and does the following while synchronized
   *           on the returned object: prints the exception by calling the UI's
   *           {@link SensibleUi#error(String, Object...) error(String,
   *           Object...)} method with an empty string as the first argument
   *           and the exception as the second argument and flushes the UI by
   *           calling the UI's {@link SensibleUi#flush() flush()} method.
   *
   * @throws StandardIoException {@inheritDoc}
   */
  @Override
  public void printError(Throwable t) {
    if (indicatesInterrupted(t)) return;
    synchronized (lock()) {
      ui.error("", t);
      ui.flush();
    }
  }

  /**
   * {@inheritDoc}
   *
   * @implSpec This implementation calls {@link #lock() lock()} and does the
   *           following while synchronized on the returned object: flushes the
   *           UI by calling the UI's {@link SensibleUi#flush() flush()}
   *           method.
   *
   * @throws StandardIoException {@inheritDoc}
   */
  @Override
  public void flush() {
    synchronized (lock()) {
      ui.flush();
    }
  }

  /**
   * Returns the user interface of this program.
   *
   * @implSpec This implementation calls {@link #lock() lock()} and does the
   *           following while synchronized on the returned object: returns
   *           this program's UI.
   *
   * @return the UI
   */
  protected T ui() {
    synchronized (lock()) {
      return ui;
    }
  }

  /**
   * Sets the user interface of this program.
   *
   * @implSpec This implementation calls {@link #lock() lock()} and does the
   *           following while synchronized on the returned object: sets this
   *           program's UI.
   *
   * @param ui the user interface of this program; this program assumes
   *          ownership of the UI; the client should not maintain references to
   *          it and must not use or modify it in any way
   *
   * @throws NullPointerException if {@code ui} is {@code null}
   */
  protected void ui(T ui) {
    synchronized (lock()) {
      this.ui = Objects.requireNonNull(ui, "ui is null");
    }
  }

  private static boolean indicatesInterrupted(Throwable t) {
    if (t instanceof InterruptedException) return true;
    if (t instanceof InterruptedIOException) return true;
    if (t instanceof ClosedByInterruptException) return true;
    return false;
  }
}
