/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import java.io.Closeable;
import java.io.Flushable;
import java.io.IOException;
import java.io.Writer;
import java.nio.CharBuffer;

/**
 * An <i>adapter</i> that adapts an {@link Appendable} to a {@link Writer}.
 * <p>
 * This class is thread-safe.
 */
public final class AppendableWriter extends Writer {
  private final Appendable adaptee;
  private boolean closed;

  public AppendableWriter(Appendable adaptee) {
    super();
    this.adaptee = adaptee;
    closed = false;
  }

  public AppendableWriter(Appendable adaptee, Object lock) {
    super(lock);
    this.adaptee = adaptee;
    closed = false;
  }

  @Override
  public AppendableWriter append(CharSequence csq) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(csq);
    }
    return this;
  }

  @Override
  public AppendableWriter append(CharSequence csq, int start, int end) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(csq, start, end);
    }
    return this;
  }

  @Override
  public AppendableWriter append(char c) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(c);
    }
    return this;
  }

  @Override
  public void write(int c) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append((char)c);
    }
  }

  @Override
  public void write(char[] cbuf) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(CharBuffer.wrap(cbuf));
    }
  }

  @Override
  public void write(char[] cbuf, int off, int len) throws IOException {
    if (off < 0) {
      throw new IndexOutOfBoundsException("off (" + off + ") < 0");
    }
    if (len < 0) {
      throw new IndexOutOfBoundsException("len (" + len + ") < 0");
    }
    int end = off + len;
    if (end < 0) {
      throw new IndexOutOfBoundsException("off + len (" + end + ") < 0");
    }
    if (end > cbuf.length) {
      throw new IndexOutOfBoundsException(
          "off + len (" + end + ") > cbuf.length (" + cbuf.length + ")");
    }
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(CharBuffer.wrap(cbuf, off, len));
    }
  }

  @Override
  public void write(String str) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(str);
    }
  }

  @Override
  public void write(String str, int off, int len) throws IOException {
    if (off < 0) {
      throw new IndexOutOfBoundsException("off (" + off + ") < 0");
    }
    if (len < 0) {
      throw new IndexOutOfBoundsException("len (" + len + ") < 0");
    }
    int end = off + len;
    if (end < 0) {
      throw new IndexOutOfBoundsException("off + len (" + end + ") < 0");
    }
    if (end > str.length()) {
      throw new IndexOutOfBoundsException(
          "off + len (" + end + ") > str.length() (" + str.length() + ")");
    }
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      adaptee.append(str, off, end);
    }
  }

  @Override
  public void flush() throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (adaptee instanceof Flushable) ((Flushable)adaptee).flush();
    }
  }

  @Override
  public void close() throws IOException {
    synchronized (lock) {
      if (closed) return;
      flush();
      if (adaptee instanceof Closeable) ((Closeable)adaptee).close();
      closed = true;
    }
  }
}
