/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import java.io.IOException;
import java.io.Writer;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.WritableByteChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import java.nio.charset.MalformedInputException;
import java.nio.charset.UnmappableCharacterException;
import java.util.Objects;

/**
 * An unbuffered {@link WritableByteChannel} {@link Writer}.
 * <p>
 * This class is not thread-safe.
 */
public final class UnbufferedWritableByteChannelWriter extends Writer {
  private final WritableByteChannel channel;
  private final CharsetEncoder encoder;
  private final ByteBuffer encoded;
  private final CharBuffer oneCharBuffer;
  private boolean closed;

  public UnbufferedWritableByteChannelWriter(WritableByteChannel channel) {
    this(channel, Charset.defaultCharset().newEncoder());
  }

  public UnbufferedWritableByteChannelWriter(WritableByteChannel channel, CharsetEncoder encoder) {
    this(channel, encoder, ByteBuffer.allocate(Utilities.defaultIOBufferSizeInBytes()));
  }

  public UnbufferedWritableByteChannelWriter(WritableByteChannel channel, CharsetEncoder encoder,
      ByteBuffer encoded) {
    super();
    this.channel = Objects.requireNonNull(channel, "channel is null");
    this.encoder = Objects.requireNonNull(encoder, "encoder is null");
    this.encoded = Objects.requireNonNull(encoded, "encoded is null");
    oneCharBuffer = CharBuffer.allocate(1);
    closed = false;
  }

  @Override
  public void write(int c) throws IOException {
    if (closed) throw new IOException("Closed");
    oneCharBuffer.clear();
    oneCharBuffer.put((char)c);
    oneCharBuffer.flip();
    write(oneCharBuffer);
  }

  @Override
  public void write(char[] cbuf) throws IOException {
    if (closed) throw new IOException("Closed");
    write(CharBuffer.wrap(cbuf));
  }

  @Override
  public void write(char[] cbuf, int off, int len) throws IOException {
    if (closed) throw new IOException("Closed");
    write(CharBuffer.wrap(cbuf, off, len));
  }

  @Override
  public void write(String str) throws IOException {
    if (closed) throw new IOException("Closed");
    write(CharBuffer.wrap(str));
  }

  @Override
  public void write(String str, int off, int len) throws IOException {
    if (closed) throw new IOException("Closed");
    /* Overflow of "off + len" is OK since the wrap method checks it. */
    write(CharBuffer.wrap(str, off, off + len));
  }

  @Override
  public Writer append(CharSequence csq) throws IOException {
    if (closed) throw new IOException("Closed");
    write(CharBuffer.wrap(csq));
    return this;
  }

  @Override
  public Writer append(CharSequence csq, int start, int end) throws IOException {
    if (closed) throw new IOException("Closed");
    write(CharBuffer.wrap(csq, start, end));
    return this;
  }

  @Override
  public Writer append(char c) throws IOException {
    if (closed) throw new IOException("Closed");
    oneCharBuffer.clear();
    oneCharBuffer.put(c);
    oneCharBuffer.flip();
    write(oneCharBuffer);
    return this;
  }

  @Override
  public void flush() throws IOException {
    if (closed) throw new IOException("Closed");
    /* NO-OP */
  }

  @Override
  public void close() throws IOException {
    if (closed) return;
    channel.close();
    closed = true;
  }

  private void write(CharBuffer unencoded) throws IOException {
    encoder.reset();
    encoded.clear();
    EncodingOperationStep step = EncodingOperationStep.ENCODE;
    for (;;) {
      CoderResult r;
      switch (step) {
        case ENCODE:
          r = encoder.encode(unencoded, encoded, false);
          break;
        case ENCODE_END:
          r = encoder.encode(unencoded, encoded, true);
          break;
        case FLUSH:
          r = encoder.flush(encoded);
          break;
        default:
          throw new AssertionError("Unexpected step: " + step);
      }
      if (r.isUnderflow()) {
        switch (step) {
          case ENCODE:
            step = EncodingOperationStep.ENCODE_END;
            break;
          case ENCODE_END:
            step = EncodingOperationStep.FLUSH;
            break;
          case FLUSH:
            encoded.flip();
            while (encoded.hasRemaining()) {
              channel.write(encoded);
            }
            return;
          default:
            throw new AssertionError("Unexpected step: " + step);
        }
      } else if (r.isOverflow()) {
        encoded.flip();
        while (encoded.hasRemaining()) {
          channel.write(encoded);
        }
        encoded.clear();
      } else if (r.isUnmappable()) {
        throw new UnmappableCharacterException(unencoded.length());
      } else {
        throw new MalformedInputException(unencoded.length());
      }
    }
  }

  /**
   * The steps of a {@link CharsetEncoder} encoding operation.
   */
  private static enum EncodingOperationStep {
    ENCODE,
    ENCODE_END,
    FLUSH
  }
}
