/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import static org.junit.Assert.assertEquals;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.EnumSet;
import java.util.Set;

import org.junit.Test;

/**
 * Tests {@link AbstractProgram}.
 */
public final class AbstractProgramTest {
  @Test
  public void testPrintError() {
    RuntimeException e = new RuntimeException("msg1");
    FakePrintWriter pw = new FakePrintWriter();
    e.printStackTrace(pw);
    String expected = pw.written();
    StringWriter w = new StringWriter();
    AbstractProgram p = newProgram(w);
    p.printError(e);
    assertEquals(expected, w.toString());
  }

  @Test
  public void testPrintError_die() {
    ProgramDieException e = new ProgramDieException("msg1", "localizedmsg1", 1);
    StringWriter w = new StringWriter();
    AbstractProgram p = newProgram(w);
    p.printError(e);
    assertEquals(String.format("%s%n", e.getMessage()), w.toString());
  }

  @Test
  public void testPrintError_dieWithLocalizedOption() {
    ProgramDieException e = new ProgramDieException("msg1", "localizedmsg1", 1);
    StringWriter w = new StringWriter();
    AbstractProgram p = newProgram(w, EnumSet.of(AbstractProgram.Option.LOCALIZED));
    p.printError(e);
    assertEquals(String.format("%s%n", e.getLocalizedMessage()), w.toString());
  }

  @Test
  public void testPrintError_dieNoMessage() {
    ProgramDieException e = new ProgramDieException(1);
    StringWriter w = new StringWriter();
    AbstractProgram p = newProgram(w);
    p.printError(e);
    assertEquals("", w.toString());
  }

  @Test
  public void testExitStatus() {
    AbstractProgram p = newProgram();
    assertEquals(Program.EXIT_FAILURE, p.exitStatus(new RuntimeException()));
  }

  @Test
  public void testExitStatus_die() {
    AbstractProgram p = newProgram();
    assertEquals(9, p.exitStatus(new ProgramDieException(9)));
  }

  private static AbstractProgram newProgram() {
    return newProgram(new StringWriter());
  }

  private static AbstractProgram newProgram(StringWriter err) {
    return newProgram(err, EnumSet.noneOf(AbstractProgram.Option.class));
  }

  private static AbstractProgram newProgram(StringWriter err, Set<AbstractProgram.Option> options) {
    return new AbstractProgram(new Object(), new PrintWriter(err), options) {
      @Override
      public int run() throws Exception {
        return 0;
      }
    };
  }
}
