/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;

import org.imca_cat.prgm.ProgramDieException;
import org.imca_cat.prgm.DefaultConfigurableSensibleUi;
import org.imca_cat.prgm.StandardIo;
import org.junit.Before;
import org.junit.Test;

/**
 * Tests {@link DefaultConfigurableSensibleUi}.
 */
public final class DefaultConfigurableSensibleUiTest {
  private static final Pattern LINE_TERMINATOR = Pattern.compile("\\r?\\n");

  private FakeProgramConsole console;
  private FakeWriter stdout;
  private FakeWriter stderr;
  private DefaultConfigurableSensibleUi.Builder uiBuilder;
  private DefaultConfigurableSensibleUi ui;

  @Before
  public void setUp() {
    console = new FakeProgramConsole();
    stdout = new FakeWriter();
    stderr = new FakeWriter();
    uiBuilder = new DefaultConfigurableSensibleUi.Builder()
        .console(console)
        .stdout(stdout)
        .stderr(stderr)
        .locale(Locale.US);
    ui = uiBuilder.build();
  }

  @Test
  public void testWrite() {
    ui.write("msg1%n");
    assertEquals("msg1", line(stdout.written(), 0));
  }

  @Test
  public void testWrite_quietIrrelevant() {
    ui = uiBuilder.quiet(true).build();
    ui.write("msg1%n");
    assertEquals("msg1", line(stdout.written(), 0));
  }

  @Test
  public void testWrite_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).build();
    ui.write("%.2f%n", 1.01);
    assertEquals("1,01", line(stdout.written(), 0));
  }

  @Test
  public void testWrite_autoFlush() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).build();
    ui.write("msg1%n");
    assertEquals("msg1", line(stdoutSw.toString(), 0));
  }

  @Test
  public void testWrite_autoFlushOff() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).autoFlushStdout(false).build();
    ui.write("msg1%n");
    assertEquals("", stdoutSw.toString());
  }

  @Test
  public void testWrite_autoFlushOffWithoutConsole() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.console(null).stdout(myStdout).build();
    ui.write("msg1%n");
    assertEquals("", stdoutSw.toString());
  }

  @Test
  public void testWrite_ioException() {
    IOException cause = new IOException();
    stdout.ioException(cause);
    try {
      ui.write("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.OUTPUT, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testWrite_ioExceptionClearedAfterThrown() {
    stdout.ioException(new IOException());
    try {
      ui.write("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stdout.ioException(null);
    ui.write("msg2%n");
  }

  @Test
  public void testWrite_ioExceptionWhenSuppressedByStreamObject() {
    stdout.ioException(new IOException());
    PrintWriter myStdout = new PrintWriter(stdout, false);
    ui = uiBuilder.stdout(myStdout).build();
    /*
     * The PrintWriter given to the UI as its standard output stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.write("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stdout.written());
    assertTrue(myStdout.checkError());
  }

  @Test
  public void testWriteStderr() {
    ui.writeStderr("msg1%n");
    assertEquals("msg1", line(stderr.written(), 0));
  }

  @Test
  public void testWriteStderr_quietIrrelevant() {
    ui = uiBuilder.quiet(true).build();
    ui.writeStderr("msg1%n");
    assertEquals("msg1", line(stderr.written(), 0));
  }

  @Test
  public void testWriteStderr_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).build();
    ui.writeStderr("%.2f%n", 1.01);
    assertEquals("1,01", line(stderr.written(), 0));
  }

  @Test
  public void testWriteStderr_autoFlush() {
    StringWriter stderrSw = new StringWriter();
    BufferedWriter myStderr = new BufferedWriter(stderrSw, 8);
    ui = uiBuilder.stderr(myStderr).autoFlushStderr(true).build();
    ui.writeStderr("msg1%n");
    assertEquals("msg1", line(stderrSw.toString(), 0));
  }

  @Test
  public void testWriteStderr_autoFlushOff() {
    StringWriter stderrSw = new StringWriter();
    BufferedWriter myStderr = new BufferedWriter(stderrSw, 8);
    ui = uiBuilder.stderr(myStderr).build();
    ui.writeStderr("msg1%n");
    assertEquals("", stderrSw.toString());
  }

  @Test
  public void testWriteStderr_ioException() {
    IOException cause = new IOException();
    stderr.ioException(cause);
    try {
      ui.writeStderr("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.ERROR, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testWriteStderr_ioExceptionClearedAfterThrown() {
    stderr.ioException(new IOException());
    try {
      ui.writeStderr("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stderr.ioException(null);
    ui.writeStderr("msg2%n");
  }

  @Test
  public void testWriteStderr_ioExceptionWhenSuppressedByStreamObject() {
    stderr.ioException(new IOException());
    PrintWriter myStderr = new PrintWriter(stderr, false);
    ui = uiBuilder.stderr(myStderr).build();
    /*
     * The PrintWriter given to the UI as its standard error stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.writeStderr("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stderr.written());
    assertTrue(myStderr.checkError());
  }

  @Test
  public void testStatus_quiet() {
    ui = uiBuilder.quiet(true).build();
    ui.status("status1%n");
    assertNull(line(stdout.written(), 0));
  }

  @Test
  public void testStatus_notQuiet() {
    ui.status("status1%n");
    assertEquals("status1", line(stdout.written(), 0));
  }

  @Test
  public void testStatus_quietCanceledOutByVerbose() {
    ui = uiBuilder.quiet(true).verbose(true).build();
    ui.status("status1%n");
    assertEquals("status1", line(stdout.written(), 0));
  }

  @Test
  public void testStatus_quietDisabledByDebug() {
    ui = uiBuilder.quiet(true).debug(true).build();
    ui.status("status1%n");
    assertEquals("status1", line(stdout.written(), 0));
  }

  @Test
  public void testStatus_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).build();
    ui.status("%.2f%n", 1.01);
    assertEquals("1,01", line(stdout.written(), 0));
  }

  @Test
  public void testStatus_toStderr() {
    ui = uiBuilder.statusTo(StandardIo.ERROR).build();
    ui.status("status1%n");
    assertEquals("status1", line(stderr.written(), 0));
  }

  @Test(expected = IllegalArgumentException.class)
  public void testStatus_toStdin() {
    uiBuilder.statusTo(StandardIo.INPUT);
  }

  @Test
  public void testStatus_prefix() {
    ui = uiBuilder.statusPrefix("status: ").build();
    ui.status("line1%n");
    assertEquals("status: line1", line(stdout.written(), 0));
  }

  @Test(expected = NullPointerException.class)
  public void testStatus_nullPrefix() {
    uiBuilder.statusPrefix(null);
  }

  @Test
  public void testStatus_autoFlush() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).build();
    ui.status("msg1%n");
    assertEquals("msg1", line(stdoutSw.toString(), 0));
  }

  @Test
  public void testStatus_autoFlushOff() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).autoFlushStdout(false).build();
    ui.status("msg1%n");
    assertEquals("", stdoutSw.toString());
  }

  @Test
  public void testStatus_ioException() {
    IOException cause = new IOException();
    stdout.ioException(cause);
    try {
      ui.status("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.OUTPUT, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testStatus_ioExceptionClearedAfterThrown() {
    stdout.ioException(new IOException());
    try {
      ui.status("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stdout.ioException(null);
    ui.status("msg2%n");
  }

  @Test
  public void testStatus_ioExceptionWhenSuppressedByStreamObject() {
    stdout.ioException(new IOException());
    PrintWriter myStdout = new PrintWriter(stdout, false);
    ui = uiBuilder.stdout(myStdout).build();
    /*
     * The PrintWriter given to the UI as its standard output stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.status("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stdout.written());
    assertTrue(myStdout.checkError());
  }

  @Test
  public void testNote_verbose() {
    ui = uiBuilder.verbose(true).build();
    ui.note("note1%n");
    assertEquals("note1", line(stdout.written(), 0));
  }

  @Test
  public void testNote_notVerbose() {
    ui.note("note1%n");
    assertNull(line(stdout.written(), 0));
  }

  @Test
  public void testNote_verboseCanceledOutByQuiet() {
    ui = uiBuilder.quiet(true).verbose(true).build();
    ui.note("note1%n");
    assertNull(line(stdout.written(), 0));
  }

  @Test
  public void testNote_verboseEnabledByDebug() {
    ui = uiBuilder.debug(true).build();
    ui.note("note1%n");
    assertEquals("note1", line(stdout.written(), 0));
  }

  @Test
  public void testNote_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).verbose(true).build();
    ui.note("%.2f%n", 1.01);
    assertEquals("1,01", line(stdout.written(), 0));
  }

  @Test
  public void testNote_toStderr() {
    ui = uiBuilder.verbose(true).noteTo(StandardIo.ERROR).build();
    ui.note("note1%n");
    assertEquals("note1", line(stderr.written(), 0));
  }

  @Test(expected = IllegalArgumentException.class)
  public void testNote_toStdin() {
    uiBuilder.noteTo(StandardIo.INPUT);
  }

  @Test
  public void testNote_prefix() {
    ui = uiBuilder.verbose(true).notePrefix("note: ").build();
    ui.note("line1%n");
    assertEquals("note: line1", line(stdout.written(), 0));
  }

  @Test(expected = NullPointerException.class)
  public void testNote_nullPrefix() {
    uiBuilder.notePrefix(null);
  }

  @Test
  public void testNote_autoFlush() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).verbose(true).build();
    ui.note("msg1%n");
    assertEquals("msg1", line(stdoutSw.toString(), 0));
  }

  @Test
  public void testNote_autoFlushOff() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).verbose(true).autoFlushStdout(false).build();
    ui.note("msg1%n");
    assertEquals("", stdoutSw.toString());
  }

  @Test
  public void testNote_ioException() {
    IOException cause = new IOException();
    stdout.ioException(cause);
    ui = uiBuilder.verbose(true).build();
    try {
      ui.note("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.OUTPUT, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testNote_ioExceptionClearedAfterThrown() {
    stdout.ioException(new IOException());
    ui = uiBuilder.verbose(true).build();
    try {
      ui.note("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stdout.ioException(null);
    ui.note("msg2%n");
  }

  @Test
  public void testNote_ioExceptionWhenSuppressedByStreamObject() {
    stdout.ioException(new IOException());
    PrintWriter myStdout = new PrintWriter(stdout, false);
    ui = uiBuilder.stdout(myStdout).verbose(true).build();
    /*
     * The PrintWriter given to the UI as its standard output stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.note("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stdout.written());
    assertTrue(myStdout.checkError());
  }

  @Test
  public void testDebug_debug() {
    ui = uiBuilder.debug(true).build();
    ui.debug("debug1%n");
    assertEquals("debug1", line(stdout.written(), 0));
  }

  @Test
  public void testDebug_notDebug() {
    ui.debug("debug1%n");
    assertNull(line(stdout.written(), 0));
  }

  @Test
  public void testDebug_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).debug(true).build();
    ui.debug("%.2f%n", 1.01);
    assertEquals("1,01", line(stdout.written(), 0));
  }

  @Test
  public void testDebug_toStderr() {
    ui = uiBuilder.debug(true).debugTo(StandardIo.ERROR).build();
    ui.debug("debug1%n");
    assertEquals("debug1", line(stderr.written(), 0));
  }

  @Test(expected = IllegalArgumentException.class)
  public void testDebug_toStdin() {
    uiBuilder.debugTo(StandardIo.INPUT);
  }

  @Test
  public void testDebug_prefix() {
    ui = uiBuilder.debug(true).debugPrefix("debug: ").build();
    ui.debug("line1%n");
    assertEquals("debug: line1", line(stdout.written(), 0));
  }

  @Test(expected = NullPointerException.class)
  public void testDebug_nullPrefix() {
    uiBuilder.debugPrefix(null);
  }

  @Test
  public void testDebug_autoFlush() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).debug(true).build();
    ui.debug("msg1%n");
    assertEquals("msg1", line(stdoutSw.toString(), 0));
  }

  @Test
  public void testDebug_autoFlushOff() {
    StringWriter stdoutSw = new StringWriter();
    BufferedWriter myStdout = new BufferedWriter(stdoutSw, 8);
    ui = uiBuilder.stdout(myStdout).debug(true).autoFlushStdout(false).build();
    ui.debug("msg1%n");
    assertEquals("", stdoutSw.toString());
  }

  @Test
  public void testDebug_ioException() {
    IOException cause = new IOException();
    stdout.ioException(cause);
    ui = uiBuilder.debug(true).build();
    try {
      ui.debug("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.OUTPUT, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testDebug_ioExceptionClearedAfterThrown() {
    stdout.ioException(new IOException());
    ui = uiBuilder.debug(true).build();
    try {
      ui.debug("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stdout.ioException(null);
    ui.debug("msg2%n");
  }

  @Test
  public void testDebug_ioExceptionWhenSuppressedByStreamObject() {
    stdout.ioException(new IOException());
    PrintWriter myStdout = new PrintWriter(stdout, false);
    ui = uiBuilder.stdout(myStdout).debug(true).build();
    /*
     * The PrintWriter given to the UI as its standard output stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.debug("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stdout.written());
    assertTrue(myStdout.checkError());
  }

  @Test
  public void testWarn() {
    ui.warn("warn1%n");
    assertEquals("warn1", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).build();
    ui.warn("%.2f%n", 1.01);
    assertEquals("1,01", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_prefix() {
    ui = uiBuilder.warnPrefix("warn: ").build();
    ui.warn("line1%n");
    assertEquals("warn: line1", line(stderr.written(), 0));
  }

  @Test(expected = NullPointerException.class)
  public void testWarn_nullPrefix() {
    uiBuilder.warnPrefix(null);
  }

  @Test
  public void testWarn_prefixFromProgramName() {
    ui = uiBuilder.programName("my-prog").build();
    ui.warn("line1%n");
    assertEquals("my-prog: line1", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_printStackTrace() {
    Exception e = new RuntimeException("msg1");
    ui.warn("", e);
    assertEquals(stackTrace(e), stderr.written());
  }

  @Test
  public void testWarn_noPrintStackTrace() {
    ui = uiBuilder.printStackTrace(false).build();
    Exception e = new RuntimeException("msg1");
    /*
     * The stack trace output format is Java platform implementation
     * dependent, but in order to determine from the output when a stack
     * trace was printed, the stack trace output format needs to at least
     * have more than one line.
     */
    assertTrue(lines(stackTrace(e)).size() > 1);
    ui.warn("", e);
    List<String> lines = lines(stderr.written());
    assertEquals(1, lines.size());
    assertEquals(e.getClass().getName() + ": msg1", lines.get(0));
  }

  @Test
  public void testWarn_useLocalizedExceptionMessage() {
    ui = uiBuilder.printStackTrace(false).useLocalizedExceptionMessage(true).build();
    @SuppressWarnings("serial")
    Exception e = new RuntimeException() {
      @Override
      public String getMessage() {
        return "non-localized msg1";
      }

      @Override
      public String getLocalizedMessage() {
        return "localized msg1";
      }
    };
    ui.warn("", e);
    assertEquals(e.getClass().getName() + ": localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_noUseLocalizedExceptionMessage() {
    ui = uiBuilder.printStackTrace(false).build();
    @SuppressWarnings("serial")
    Exception e = new RuntimeException() {
      @Override
      public String getMessage() {
        return "non-localized msg1";
      }

      @Override
      public String getLocalizedMessage() {
        return "localized msg1";
      }
    };
    ui.warn("", e);
    assertEquals(e.getClass().getName() + ": non-localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_dieNoStackTraceEvenIfEnabled() {
    Exception e = new ProgramDieException("msg1", 1);
    /*
     * The stack trace output format is Java platform implementation
     * dependent, but in order to determine from the output when a stack
     * trace was printed, the stack trace output format needs to at least
     * have more than one line.
     */
    assertTrue(lines(stackTrace(e)).size() > 1);
    ui.warn("", e);
    assertEquals(1, lines(stderr.written()).size());
  }

  @Test
  public void testWarn_dieUseLocalizedExceptionMessage() {
    ui = uiBuilder.useLocalizedExceptionMessage(true).build();
    ui.warn("", new ProgramDieException("non-localized msg1", "localized msg1", 1));
    assertEquals("localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_dieNoUseLocalizedExceptionMessage() {
    ui.warn("", new ProgramDieException("non-localized msg1", "localized msg1", 1));
    assertEquals("non-localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testWarn_dieNoMessage() {
    ui.warn("", new ProgramDieException(1));
    assertTrue(lines(stderr.written()).isEmpty());
  }

  @Test
  public void testWarn_dieNoMessageWithPrefix() {
    ui = uiBuilder.warnPrefix("warn: ").build();
    ui.warn("", new ProgramDieException(1));
    assertTrue(lines(stderr.written()).isEmpty());
  }

  @Test
  public void testWarn_autoFlush() {
    StringWriter stderrSw = new StringWriter();
    BufferedWriter myStderr = new BufferedWriter(stderrSw, 8);
    ui = uiBuilder.stderr(myStderr).autoFlushStderr(true).build();
    ui.warn("msg1%n");
    assertEquals("msg1", line(stderrSw.toString(), 0));
  }

  @Test
  public void testWarn_autoFlushOff() {
    StringWriter stderrSw = new StringWriter();
    BufferedWriter myStderr = new BufferedWriter(stderrSw, 8);
    ui = uiBuilder.stderr(myStderr).build();
    ui.warn("msg1%n");
    assertEquals("", stderrSw.toString());
  }

  @Test
  public void testWarn_ioException() {
    IOException cause = new IOException();
    stderr.ioException(cause);
    try {
      ui.warn("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.ERROR, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testWarn_ioExceptionClearedAfterThrown() {
    stderr.ioException(new IOException());
    try {
      ui.warn("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stderr.ioException(null);
    ui.warn("msg2%n");
  }

  @Test
  public void testWarn_ioExceptionWhenSuppressedByStreamObject() {
    stderr.ioException(new IOException());
    PrintWriter myStderr = new PrintWriter(stderr, false);
    ui = uiBuilder.stderr(myStderr).build();
    /*
     * The PrintWriter given to the UI as its standard error stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.warn("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stderr.written());
    assertTrue(myStderr.checkError());
  }

  @Test
  public void testWarn_ioExceptionWithStackTraceWhenSuppressedByStreamObject() {
    Exception e = new RuntimeException("msg1");
    stderr.ioException(new IOException());
    PrintWriter myStderr = new PrintWriter(stderr, false);
    ui = uiBuilder.stderr(myStderr).build();
    /*
     * The PrintWriter given to the UI as its standard error stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.warn("", e);
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stderr.written());
    assertTrue(myStderr.checkError());
  }

  @Test
  public void testError() {
    ui.error("error1%n");
    assertEquals("error1", line(stderr.written(), 0));
  }

  @Test
  public void testError_locale() {
    ui = uiBuilder.locale(Locale.FRANCE).build();
    ui.error("%.2f%n", 1.01);
    assertEquals("1,01", line(stderr.written(), 0));
  }

  @Test
  public void testError_prefix() {
    ui = uiBuilder.errorPrefix("error: ").build();
    ui.error("line1%n");
    assertEquals("error: line1", line(stderr.written(), 0));
  }

  @Test(expected = NullPointerException.class)
  public void testError_nullPrefix() {
    uiBuilder.errorPrefix(null);
  }

  @Test
  public void testError_prefixFromProgramName() {
    ui = uiBuilder.programName("my-prog").build();
    ui.error("line1%n");
    assertEquals("my-prog: line1", line(stderr.written(), 0));
  }

  @Test
  public void testError_printStackTrace() {
    Exception e = new RuntimeException("msg1");
    ui.error("", e);
    assertEquals(stackTrace(e), stderr.written());
  }

  @Test
  public void testError_noPrintStackTrace() {
    ui = uiBuilder.printStackTrace(false).build();
    Exception e = new RuntimeException("msg1");
    /*
     * The stack trace output format is Java platform implementation
     * dependent, but in order to determine from the output when a stack
     * trace was printed, the stack trace output format needs to at least
     * have more than one line.
     */
    assertTrue(lines(stackTrace(e)).size() > 1);
    ui.error("", e);
    List<String> lines = lines(stderr.written());
    assertEquals(1, lines.size());
    assertEquals(e.getClass().getName() + ": msg1", lines.get(0));
  }

  @Test
  public void testError_useLocalizedExceptionMessage() {
    ui = uiBuilder.printStackTrace(false).useLocalizedExceptionMessage(true).build();
    @SuppressWarnings("serial")
    Exception e = new RuntimeException() {
      @Override
      public String getMessage() {
        return "non-localized msg1";
      }

      @Override
      public String getLocalizedMessage() {
        return "localized msg1";
      }
    };
    ui.error("", e);
    assertEquals(e.getClass().getName() + ": localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testError_noUseLocalizedExceptionMessage() {
    ui = uiBuilder.printStackTrace(false).build();
    @SuppressWarnings("serial")
    Exception e = new RuntimeException() {
      @Override
      public String getMessage() {
        return "non-localized msg1";
      }

      @Override
      public String getLocalizedMessage() {
        return "localized msg1";
      }
    };
    ui.error("", e);
    assertEquals(e.getClass().getName() + ": non-localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testError_dieNoStackTraceEvenIfEnabled() {
    Exception e = new ProgramDieException("msg1", 1);
    /*
     * The stack trace output format is Java platform implementation
     * dependent, but in order to determine from the output when a stack
     * trace was printed, the stack trace output format needs to at least
     * have more than one line.
     */
    assertTrue(lines(stackTrace(e)).size() > 1);
    ui.error("", e);
    assertEquals(1, lines(stderr.written()).size());
  }

  @Test
  public void testError_dieUseLocalizedExceptionMessage() {
    ui = uiBuilder.useLocalizedExceptionMessage(true).build();
    ui.error("", new ProgramDieException("non-localized msg1", "localized msg1", 1));
    assertEquals("localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testError_dieNoUseLocalizedExceptionMessage() {
    ui.error("", new ProgramDieException("non-localized msg1", "localized msg1", 1));
    assertEquals("non-localized msg1", line(stderr.written(), 0));
  }

  @Test
  public void testError_dieNoMessage() {
    ui.error("", new ProgramDieException(1));
    assertTrue(lines(stderr.written()).isEmpty());
  }

  @Test
  public void testError_dieNoMessageWithPrefix() {
    ui = uiBuilder.errorPrefix("error: ").build();
    ui.error("", new ProgramDieException(1));
    assertTrue(lines(stderr.written()).isEmpty());
  }

  @Test
  public void testError_autoFlush() {
    StringWriter stderrSw = new StringWriter();
    BufferedWriter myStderr = new BufferedWriter(stderrSw, 8);
    ui = uiBuilder.stderr(myStderr).autoFlushStderr(true).build();
    ui.error("msg1%n");
    assertEquals("msg1", line(stderrSw.toString(), 0));
  }

  @Test
  public void testError_autoFlushOff() {
    StringWriter stderrSw = new StringWriter();
    BufferedWriter myStderr = new BufferedWriter(stderrSw, 8);
    ui = uiBuilder.stderr(myStderr).build();
    ui.error("msg1%n");
    assertEquals("", stderrSw.toString());
  }

  @Test
  public void testError_ioException() {
    IOException cause = new IOException();
    stderr.ioException(cause);
    try {
      ui.error("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      assertEquals(StandardIo.ERROR, e.io());
      assertEquals(cause, e.getCause());
    }
  }

  @Test
  public void testError_ioExceptionClearedAfterThrown() {
    stderr.ioException(new IOException());
    try {
      ui.error("msg1%n");
      fail("Expected StandardIoException");
    } catch (StandardIoException e) {
      /* EXPECTED */
    }
    stderr.ioException(null);
    ui.error("msg2%n");
  }

  @Test
  public void testError_ioExceptionWhenSuppressedByStreamObject() {
    stderr.ioException(new IOException());
    PrintWriter myStderr = new PrintWriter(stderr, false);
    ui = uiBuilder.stderr(myStderr).build();
    /*
     * The PrintWriter given to the UI as its standard error stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.error("msg1%n");
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stderr.written());
    assertTrue(myStderr.checkError());
  }

  @Test
  public void testError_ioExceptionWithStackTraceWhenSuppressedByStreamObject() {
    Exception e = new RuntimeException("msg1");
    stderr.ioException(new IOException());
    PrintWriter myStderr = new PrintWriter(stderr, false);
    ui = uiBuilder.stderr(myStderr).build();
    /*
     * The PrintWriter given to the UI as its standard error stream will
     * catch and store the IOException thrown by the underlying stream, so
     * the UI will not see it.
     */
    ui.error("", e);
    /* Because an IOException was thrown, the message was not written. */
    assertEquals("", stderr.written());
    assertTrue(myStderr.checkError());
  }

  @Test
  public void testFlush() {
    StringWriter stdoutSw = new StringWriter();
    PrintWriter myStdout = new PrintWriter(new BufferedWriter(stdoutSw, 8), false);
    StringWriter stderrSw = new StringWriter();
    PrintWriter myStderr = new PrintWriter(new BufferedWriter(stderrSw, 8), false);
    StringWriter consoleSw = new StringWriter();
    PrintWriter consolePw = new PrintWriter(new BufferedWriter(consoleSw, 8), false);
    BufferedReader consoleBr = new BufferedReader(new StringReader(""));
    ProgramConsole myConsole = new FakeSynchronizedProgramConsole(consolePw, consoleBr);
    ui = uiBuilder
        .console(myConsole)
        .stdout(myStdout)
        .stderr(myStderr)
        .autoFlushStdout(false)
        .build();
    ui.write("msg1%n");
    ui.writeStderr("msg2%n");
    myConsole.format("msg3%n");
    assertEquals("", stdoutSw.toString());
    assertEquals("", stderrSw.toString());
    assertEquals("", consoleSw.toString());
    ui.flush();
    assertEquals("msg1", line(stdoutSw.toString(), 0));
    assertEquals("msg2", line(stderrSw.toString(), 0));
    assertEquals("msg3", line(consoleSw.toString(), 0));
  }

  @Test
  public void testPrompt() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("Anakin\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    String name = ui.prompt("What is your name?", "Vader");
    assertEquals("What is your name? ", consoleW.written());
    assertEquals("Anakin", name);
  }

  @Test
  public void testPrompt_defaultDueToEnter() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    String name = ui.prompt("What is your name?", "Vader");
    assertEquals("Vader", name);
  }

  @Test
  public void testPrompt_defaultDueToEndOfStream() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    String name = ui.prompt("What is your name?", "Vader");
    assertEquals("Vader", name);
  }

  @Test(expected = NullPointerException.class)
  public void testPrompt_nullMessage() {
    ui.prompt(null, "Anakin");
  }

  @Test
  public void testPromptChoice() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("n\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    int choice = ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "y,n");
    assertEquals("Shut down all garbage mashers (Yn)? ", consoleW.written());
    assertEquals(1, choice);
  }

  @Test
  public void testPromptChoice_uppercaseUserInput() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("N\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    int choice = ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "y,n");
    assertEquals(1, choice);
  }

  @Test
  public void testPromptChoice_alternateUserInput() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("NO\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    int choice = ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "y/yes,n/no");
    assertEquals(1, choice);
  }

  @Test
  public void testPromptChoice_defaultDueToEnter() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("\n\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    int choice = ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "y,n");
    assertEquals(0, choice);
    /* Default is not first index. */
    choice = ui.promptChoice("Shut down all garbage mashers (yN)?", 1, "y,n");
    assertEquals(1, choice);
  }

  @Test
  public void testPromptChoice_defaultDueToEndOfStream() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    int choice = ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "y,n");
    assertEquals(0, choice);
  }

  @Test
  public void testPromptChoice_noMatch() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("x\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    int choice = ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "y,n");
    assertEquals(-1, choice);
  }

  @Test(expected = NullPointerException.class)
  public void testPromptChoice_nullMessage() {
    ui.promptChoice(null, 0, "y,n");
  }

  @Test(expected = IllegalArgumentException.class)
  public void testPromptChoice_defaultLessThanZero() {
    ui.promptChoice("Shut down all garbage mashers (Yn)?", -1, "y,n");
  }

  @Test(expected = NullPointerException.class)
  public void testPromptChoice_nullChoices() {
    ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, null);
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testPromptChoice_badChoices() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("n\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    ui.promptChoice("Shut down all garbage mashers (Yn)?", 0, "");
  }

  @Test
  public void testPromptPassword() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("whatever\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    char[] password = ui.promptPassword("Password:", "12345".toCharArray());
    assertEquals("Password: ", consoleW.written());
    assertArrayEquals("whatever".toCharArray(), password);
  }

  @Test
  public void testPromptPassword_defaultDueToEnter() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("\n", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    char[] password = ui.promptPassword("Password:", "12345".toCharArray());
    assertArrayEquals("12345".toCharArray(), password);
  }

  @Test
  public void testPromptPassword_defaultDueToEndOfStream() {
    Object lock = new Object();
    FakePrintWriter consoleW = new FakePrintWriter(true, Locale.US, lock);
    FakeReader consoleR = new FakeReader("", lock);
    console = new FakeProgramConsole(consoleW, consoleR, lock);
    ui = uiBuilder.console(console).build();
    char[] password = ui.promptPassword("Password:", "12345".toCharArray());
    assertArrayEquals("12345".toCharArray(), password);
  }

  @Test(expected = NullPointerException.class)
  public void testPromptPassword_nullMessage() {
    ui.promptPassword(null, "whatever".toCharArray());
  }

  @Test
  public void testToBuilder_prefixFromProgramName() {
    ui = uiBuilder.programName("my-prog").build();
    ui.warn("line1%n");
    ui.error("line2%n");
    assertEquals("my-prog: line1", line(stderr.written(), 0));
    assertEquals("my-prog: line2", line(stderr.written(), 1));
    DefaultConfigurableSensibleUi uix = ui.toBuilder().programName("my-progx").build();
    uix.warn("line1%n");
    uix.error("line2%n");
    assertEquals("my-progx: line1", line(stderr.written(), 2));
    assertEquals("my-progx: line2", line(stderr.written(), 3));
  }

  private static String line(String s, int index) {
    List<String> lines = lines(s);
    return (index >= 0 && index < lines.size()) ? lines.get(index) : null;
  }

  private static List<String> lines(String s) {
    if (s.isEmpty()) return Collections.emptyList();
    List<String> result = Arrays.asList(LINE_TERMINATOR.split(s, -1));
    if (result.size() == 1) return result;
    if (result.get(result.size() - 1).isEmpty()) return result.subList(0, result.size() - 1);
    return result;
  }

  private static String stackTrace(Throwable t) {
    try (StringWriter sw = new StringWriter(); PrintWriter pw = new PrintWriter(sw)) {
      t.printStackTrace(pw);
      pw.flush();
      return sw.toString();
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }
}
