/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.PrintWriter;
import java.util.Locale;

/**
 * A Fake that writes into memory.
 * <p>
 * What has been written can be retrieved via the {@code written()} method.
 * <p>
 * This class is thread-safe.
 * <p>
 * This class is unfortunately necessary because the {@code PrintWriter}
 * documentation does not specify whether it is thread-safe. Also,
 * {@code PrintWriter} does not provide a constructor that takes a lock object
 * as an argument. Lastly, the abstract {@code Writer} class provides a
 * {@code lock} instance field, but it does not specify under what conditions,
 * if any, it can be set by a subclass.
 */
public final class FakePrintWriter extends PrintWriter {
  private final Locale locale;
  private final Object mutex;

  public FakePrintWriter() {
    this(false, new Object());
  }

  public FakePrintWriter(boolean autoFlush, Object lock) {
    this(autoFlush, Locale.getDefault(), lock);
  }

  public FakePrintWriter(boolean autoFlush, Locale locale, Object lock) {
    super(new FakeWriter(lock), autoFlush);
    this.locale = locale;
    this.mutex = lock;
  }

  @Override
  public void flush() {
    synchronized (mutex) {
      super.flush();
    }
  }

  @Override
  public void close() {
    synchronized (mutex) {
      super.close();
    }
  }

  @Override
  public boolean checkError() {
    synchronized (mutex) {
      return super.checkError();
    }
  }

  @Override
  public void write(int c) {
    synchronized (mutex) {
      super.write(c);
    }
  }

  @Override
  public void write(char[] buf, int off, int len) {
    synchronized (mutex) {
      super.write(buf, off, len);
    }
  }

  @Override
  public void write(char[] buf) {
    synchronized (mutex) {
      super.write(buf);
    }
  }

  @Override
  public void write(String s, int off, int len) {
    synchronized (mutex) {
      super.write(s, off, len);
    }
  }

  @Override
  public void write(String s) {
    synchronized (mutex) {
      super.write(s);
    }
  }

  @Override
  public void print(boolean b) {
    synchronized (mutex) {
      super.print(b);
    }
  }

  @Override
  public void print(char c) {
    synchronized (mutex) {
      super.print(c);
    }
  }

  @Override
  public void print(int i) {
    synchronized (mutex) {
      super.print(i);
    }
  }

  @Override
  public void print(long l) {
    synchronized (mutex) {
      super.print(l);
    }
  }

  @Override
  public void print(float f) {
    synchronized (mutex) {
      super.print(f);
    }
  }

  @Override
  public void print(double d) {
    synchronized (mutex) {
      super.print(d);
    }
  }

  @Override
  public void print(char[] s) {
    synchronized (mutex) {
      super.print(s);
    }
  }

  @Override
  public void print(String s) {
    synchronized (mutex) {
      super.print(s);
    }
  }

  @Override
  public void print(Object obj) {
    synchronized (mutex) {
      super.print(obj);
    }
  }

  @Override
  public void println() {
    synchronized (mutex) {
      super.println();
    }
  }

  @Override
  public void println(boolean x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(char x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(int x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(long x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(float x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(double x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(char[] x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(String x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public void println(Object x) {
    synchronized (mutex) {
      super.println(x);
    }
  }

  @Override
  public PrintWriter printf(String format, Object... args) {
    return printf(locale, format, args);
  }

  @Override
  public PrintWriter printf(Locale l, String format, Object... args) {
    synchronized (mutex) {
      return super.printf(l, format, args);
    }
  }

  @Override
  public PrintWriter format(String format, Object... args) {
    return format(locale, format, args);
  }

  @Override
  public PrintWriter format(Locale l, String format, Object... args) {
    synchronized (mutex) {
      return super.format(l, format, args);
    }
  }

  @Override
  public PrintWriter append(CharSequence csq) {
    synchronized (mutex) {
      return super.append(csq);
    }
  }

  @Override
  public PrintWriter append(CharSequence csq, int start, int end) {
    synchronized (mutex) {
      return super.append(csq, start, end);
    }
  }

  @Override
  public PrintWriter append(char c) {
    synchronized (mutex) {
      return super.append(c);
    }
  }

  public String written() {
    synchronized (mutex) {
      flush();
      return ((FakeWriter)out).written();
    }
  }
}
