/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.IOError;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.Locale;
import java.util.Objects;

import org.imca_cat.prgm.ProgramConsole;

/**
 * A Fake that writes to a {@link FakePrintWriter} and reads from a
 * {@link FakeReader}.
 */
public final class FakeProgramConsole implements ProgramConsole {
  private final FakePrintWriter writer;
  private final FakeReader reader;
  private final Object lock;

  public FakeProgramConsole() {
    this(Locale.getDefault());
  }

  public FakeProgramConsole(Locale locale) {
    this(locale, new Object());
  }

  private FakeProgramConsole(Locale locale, Object lock) {
    this(new FakePrintWriter(true, locale, lock), new FakeReader(lock), lock);
  }

  /**
   * The writer and reader must synchronize internally on the specified lock.
   */
  public FakeProgramConsole(FakePrintWriter writer, FakeReader reader, Object lock) {
    super();
    this.writer = Objects.requireNonNull(writer, "writer is null");
    this.reader = Objects.requireNonNull(reader, "reader is null");
    this.lock = Objects.requireNonNull(lock, "lock is null");
  }

  @Override
  public PrintWriter writer() {
    synchronized (lock) {
      return writer;
    }
  }

  @Override
  public Reader reader() {
    synchronized (lock) {
      return reader;
    }
  }

  @Override
  public FakeProgramConsole format(String fmt, Object... args) {
    synchronized (lock) {
      writer.format(fmt, args);
      return this;
    }
  }

  @Override
  public FakeProgramConsole printf(String format, Object... args) {
    synchronized (lock) {
      writer.printf(format, args);
      return this;
    }
  }

  @Override
  public String readLine(String fmt, Object... args) {
    synchronized (lock) {
      writer.format(fmt, args);
      writer.flush();
      try {
        return reader.readLine();
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public String readLine() {
    synchronized (lock) {
      try {
        return reader.readLine();
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public char[] readPassword(String fmt, Object... args) {
    synchronized (lock) {
      writer.format(fmt, args);
      writer.flush();
      try {
        String result = reader.readLine();
        return (result != null) ? result.toCharArray() : null;
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public char[] readPassword() {
    synchronized (lock) {
      try {
        String result = reader.readLine();
        return (result != null) ? result.toCharArray() : null;
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public void flush() {
    synchronized (lock) {
      writer.flush();
    }
  }

  public String written() {
    synchronized (lock) {
      return writer.written();
    }
  }
}
