/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.nio.CharBuffer;
import java.util.Objects;

/**
 * A Fake that reads from a {@code String} source.
 * <p>
 * This class is thread-safe.
 * <p>
 * This class is unfortunately necessary because the {@code BufferedReader}
 * documentation does not specify whether it is thread-safe. Also,
 * {@code BufferedReader} does not provide a constructor that takes a lock
 * object as an argument. Lastly, the abstract {@code Reader} class provides a
 * {@code lock} instance field, but it does not specify under what conditions,
 * if any, it can be set by a subclass.
 */
public final class FakeReader extends Reader {
  private final BufferedReader source;
  private boolean closed;

  public FakeReader() {
    this(new Object());
  }

  public FakeReader(Object lock) {
    this("", lock);
  }

  public FakeReader(String source, Object lock) {
    super(Objects.requireNonNull(lock, "lock is null"));
    this.source = new BufferedReader(
        new StringReader(Objects.requireNonNull(source, "source is null")));
    closed = false;
  }

  @Override
  public int read() throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.read();
    }
  }

  @Override
  public int read(char[] cbuf, int off, int len) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.read(cbuf, off, len);
    }
  }

  public String readLine() throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.readLine();
    }
  }

  @Override
  public int read(CharBuffer target) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.read(target);
    }
  }

  @Override
  public long skip(long ns) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.skip(ns);
    }
  }

  @Override
  public int read(char[] cbuf) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.read(cbuf);
    }
  }

  @Override
  public boolean ready() throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      return source.ready();
    }
  }

  @Override
  public boolean markSupported() {
    synchronized (lock) {
      return source.markSupported();
    }
  }

  @Override
  public void mark(int readAheadLimit) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      source.mark(readAheadLimit);
    }
  }

  @Override
  public void reset() throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      source.reset();
    }
  }

  @Override
  public void close() throws IOException {
    synchronized (lock) {
      if (closed) return;
      source.close();
      closed = true;
    }
  }
}
