/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.BufferedReader;
import java.io.IOError;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.Objects;

/**
 * A Fake that does not require its underlying writer and reader to be
 * thread-safe.
 * <p>
 * This class is conditionally thread-safe. All methods may be called
 * concurrently. However, the client must manually synchronize on this console
 * if it wants to call any of the methods directly on the underlying writer and
 * reader specified in the constructor and returned by {@link #writer()} and
 * {@link #reader()}, respectively.
 */
public final class FakeSynchronizedProgramConsole implements ProgramConsole {
  private final PrintWriter writer;
  private final BufferedReader reader;
  private final Object lock;

  public FakeSynchronizedProgramConsole(PrintWriter writer, BufferedReader reader) {
    super();
    this.writer = Objects.requireNonNull(writer, "writer is null");
    this.reader = Objects.requireNonNull(reader, "reader is null");
    lock = this;
  }

  @Override
  public PrintWriter writer() {
    return writer;
  }

  @Override
  public Reader reader() {
    return reader;
  }

  @Override
  public FakeSynchronizedProgramConsole format(String fmt, Object... args) {
    synchronized (lock) {
      writer.format(fmt, args);
      return this;
    }
  }

  @Override
  public FakeSynchronizedProgramConsole printf(String format, Object... args) {
    synchronized (lock) {
      writer.printf(format, args);
      return this;
    }
  }

  @Override
  public String readLine(String fmt, Object... args) {
    synchronized (lock) {
      writer.format(fmt, args);
      writer.flush();
      try {
        return reader.readLine();
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public String readLine() {
    synchronized (lock) {
      try {
        return reader.readLine();
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public char[] readPassword(String fmt, Object... args) {
    synchronized (lock) {
      writer.format(fmt, args);
      writer.flush();
      try {
        String result = reader.readLine();
        return (result != null) ? result.toCharArray() : null;
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public char[] readPassword() {
    synchronized (lock) {
      try {
        String result = reader.readLine();
        return (result != null) ? result.toCharArray() : null;
      } catch (IOException e) {
        throw new IOError(e);
      }
    }
  }

  @Override
  public void flush() {
    synchronized (lock) {
      writer.flush();
    }
  }
}
