/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm;

import java.io.IOException;
import java.io.Writer;
import java.util.Objects;

/**
 * A Fake that writes into memory.
 * <p>
 * What has been written can be retrieved via the {@code written()} method.
 * <p>
 * This class is thread-safe.
 * <p>
 * This class is unfortunately necessary because the {@code StringWriter}
 * documentation does not specify whether it is thread-safe. Also,
 * {@code StringWriter} does not provide a constructor that takes a lock object
 * as an argument. Lastly, the abstract {@code Writer} class provides a
 * {@code lock} instance field, but it does not specify under what conditions,
 * if any, it can be set by a subclass.
 */
public final class FakeWriter extends Writer {
  private final StringBuilder buffer;
  private boolean closed;
  private IOException ioException;
  private IOException closeIOException;

  public FakeWriter() {
    this(new Object());
  }

  public FakeWriter(Object lock) {
    super(Objects.requireNonNull(lock, "lock is null"));
    buffer = new StringBuilder();
    closed = false;
    ioException = null;
  }

  @Override
  public void write(int c) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append((char)c);
    }
  }

  @Override
  public void write(char[] cbuf) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(cbuf);
    }
  }

  @Override
  public void write(char[] cbuf, int off, int len) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(cbuf, off, len);
    }
  }

  @Override
  public void write(String str) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(str);
    }
  }

  @Override
  public void write(String str, int off, int len) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(str, off, len);
    }
  }

  @Override
  public FakeWriter append(CharSequence csq) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(csq);
      return this;
    }
  }

  @Override
  public FakeWriter append(CharSequence csq, int start, int end) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(csq, start, end);
      return this;
    }
  }

  @Override
  public FakeWriter append(char c) throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
      buffer.append(c);
      return this;
    }
  }

  @Override
  public void flush() throws IOException {
    synchronized (lock) {
      if (closed) throw new IOException("Closed");
      if (ioException != null) throw ioException;
    }
  }

  @Override
  public void close() throws IOException {
    synchronized (lock) {
      if (closed) return;
      if (closeIOException != null) throw closeIOException;
      closed = true;
    }
  }

  public String written() {
    synchronized (lock) {
      return buffer.toString();
    }
  }

  public void ioException(IOException ioException) {
    synchronized (lock) {
      this.ioException = ioException;
    }
  }

  public void closeIOException(IOException closeIOException) {
    synchronized (lock) {
      this.closeIOException = closeIOException;
    }
  }
}
