/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;

import org.junit.Test;

/**
 * Tests {@link AppendableWriter}.
 */
public final class AppendableWriterTest {
  @Test
  public void testAppend_char() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.append('a');
      assertEquals("a", b.toString());
    }
  }

  @Test
  public void testAppend_charSequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.append("abc");
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testAppend_charSequenceSubsequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.append("xxxxxabcx", 5, 8);
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testAppend_emptyCharSequenceSubsequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.append("xxxxxabcx", 5, 5);
      assertEquals("", b.toString());
    }
  }

  @Test
  public void testWrite_charArray() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("abc".toCharArray());
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testWrite_charArraySubsequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("xxxxxabcx".toCharArray(), 5, 3);
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testWrite_emptyCharArraySubsequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("xxxxxabcx".toCharArray(), 5, 0);
      assertEquals("", b.toString());
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_charArrayNegativeOffset() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      w.write("abc".toCharArray(), -1, 3);
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_charArrayNegativeLength() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      w.write("abc".toCharArray(), 0, -1);
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_charArrayNegativeOffsetPlusLength() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      /* Integer overflow in "offset + length" results in negative integer. */
      w.write("abc".toCharArray(), Integer.MAX_VALUE, 1);
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_charArrayOffsetPlusLengthTooBig() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      w.write("abc".toCharArray(), 0, 4);
    }
  }

  @Test
  public void testWrite_int() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write(Character.codePointAt("a", 0));
      assertEquals("a", b.toString());
    }
  }

  @Test
  public void testWrite_string() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("abc");
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testWrite_emptyString() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("");
      assertEquals("", b.toString());
    }
  }

  @Test
  public void testWrite_stringSubsequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("xxxxxabcx", 5, 3);
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testWrite_emptyStringSubsequence() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.write("xxxxxabcx", 5, 0);
      assertEquals("", b.toString());
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_stringSubsequenceNegativeOffset() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      w.write("abc", -1, 3);
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_stringSubsequenceNegativeLength() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      w.write("abc", 0, -1);
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_stringSubsequenceNegativeOffsetPlusLength() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      /* Integer overflow in "offset + length" results in negative integer. */
      w.write("abc", Integer.MAX_VALUE, 1);
    }
  }

  @Test(expected = IndexOutOfBoundsException.class)
  public void testWrite_stringSubsequenceOffsetPlusLengthTooBig() throws IOException {
    try (AppendableWriter w = new AppendableWriter(new StringBuilder())) {
      w.write("abc", 0, 4);
    }
  }

  @Test
  public void testClose() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.append("abc");
      w.close();
    }
  }

  @Test
  public void testClose_nothingWritten() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.close();
    }
  }

  @Test
  public void testClose_twice() throws IOException {
    StringBuilder b = new StringBuilder();
    try (AppendableWriter w = new AppendableWriter(b)) {
      w.append("abc");
      w.close();
      w.close();
      assertEquals("abc", b.toString());
    }
  }

  @Test
  public void testClose_flushes() throws IOException {
    StringWriter sw = new StringWriter();
    try (AppendableWriter w = new AppendableWriter(new BufferedWriter(sw, 8))) {
      w.append("abc");
      assertTrue(sw.toString().isEmpty());
      w.close();
      assertEquals("abc", sw.toString());
    }
  }

  @Test
  public void testFlush() throws IOException {
    StringWriter sw = new StringWriter();
    try (AppendableWriter w = new AppendableWriter(new BufferedWriter(sw, 8))) {
      w.append("abc");
      assertTrue(sw.toString().isEmpty());
      w.flush();
      assertEquals("abc", sw.toString());
    }
  }
}
