/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import static org.junit.Assert.assertEquals;

import java.util.Arrays;
import java.util.List;

import org.imca_cat.prgm.PromptChoicesSyntaxException;
import org.junit.Test;

/**
 * Tests {@link PromptChoicesParser}.
 */
public final class PromptChoicesParserTest {
  @Test
  public void testParse_twoChoicesWithoutAlternates() {
    String in = "y,n";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y"),
        Arrays.asList("n"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_twoChoicesLongerThanOneLetterWithoutAlternates() {
    String in = "yes,no";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("yes"),
        Arrays.asList("no"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_twoChoicesWithAlternates() {
    String in = "y/yes,n/no";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y", "yes"),
        Arrays.asList("n", "no"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_twoChoicesWithThreeAlternates() {
    String in = "y/yes/yeah,n/no/nah";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y", "yes", "yeah"),
        Arrays.asList("n", "no", "nah"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_threeChoicesWithAlternates() {
    String in = "y/yes,n/no,?/help";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y", "yes"),
        Arrays.asList("n", "no"),
        Arrays.asList("?", "help"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_escapeBackslash() {
    String in = "y\\\\/yes,n/no";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y\\", "yes"),
        Arrays.asList("n", "no"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_escapeSlash() {
    String in = "y\\/yes,n/no";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y/yes"),
        Arrays.asList("n", "no"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test
  public void testParse_escapeComma() {
    String in = "y/yes\\,,n/no";
    /* @formatter:off */
    List<List<String>> expected = Arrays.asList(
        Arrays.asList("y", "yes,"),
        Arrays.asList("n", "no"));
    /* @formatter:on */
    assertEquals(expected, PromptChoicesParser.parse(in));
  }

  @Test(expected = NullPointerException.class)
  public void testParse_null() {
    PromptChoicesParser.parse(null);
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testParse_empty() {
    PromptChoicesParser.parse("");
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testParse_onlyOneChoice() {
    PromptChoicesParser.parse("y");
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testParse_emptyAlternateBeforeSlash() {
    PromptChoicesParser.parse("/y,n");
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testParse_emptyAlternateBeforeComma() {
    PromptChoicesParser.parse("y/,n");
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testParse_emptyAlternates() {
    PromptChoicesParser.parse("y,,n");
  }

  @Test(expected = PromptChoicesSyntaxException.class)
  public void testParse_trailingEscape() {
    PromptChoicesParser.parse("y,n\\");
  }
}
