/*
 * Copyright (c) 2018 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.prgm.internal;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.StandardCharsets;

import org.junit.Test;

/**
 * Tests {@link UnbufferedWritableByteChannelWriter}.
 */
public final class UnbufferedWritableByteChannelWriterTest {
  private static final Charset UTF_8 = StandardCharsets.UTF_8;

  @Test(expected = NullPointerException.class)
  public void testConstructor_nullChannel() throws IOException {
    try (Writer w = new UnbufferedWritableByteChannelWriter(null, UTF_8.newEncoder(),
        ByteBuffer.allocate(0))) {
    }
  }

  @Test(expected = NullPointerException.class)
  public void testConstructor_nullEncoder() throws IOException {
    try (Writer w = new UnbufferedWritableByteChannelWriter(
        Channels.newChannel(new ByteArrayOutputStream(0)), null, ByteBuffer.allocate(0))) {
    }
  }

  @Test(expected = NullPointerException.class)
  public void testConstructor_nullByteBuffer() throws IOException {
    try (Writer w = new UnbufferedWritableByteChannelWriter(
        Channels.newChannel(new ByteArrayOutputStream(0)), UTF_8.newEncoder(), null)) {
    }
  }

  @Test
  public void testWriteInt() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write((int)'a');
      assertEquals("a", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteInt_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      /* Euro sign. */
      w.write(0x20_AC);
      assertEquals(3, stream.size());
      assertEquals("\u20AC", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArray() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abcdefg".toCharArray());
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArray_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abc \u20AC def".toCharArray());
      assertEquals("abc \u20AC def", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArray_encodedBufferOverflow() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      w.write("abcdefg".toCharArray());
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArray_encodedBufferOverflowOnMultiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      /* Overflows after first byte of three for euro sign. */
      w.write("abc\u20ACdef".toCharArray());
      assertEquals("abc\u20ACdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArrayRange() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abcdefg".toCharArray(), 1, 5);
      assertEquals("bcdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArrayRange_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abc \u20AC def".toCharArray(), 1, 7);
      assertEquals("bc \u20AC de", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArrayRange_all() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abcdefg".toCharArray(), 0, 7);
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteCharArrayRange_encodedBufferOverflow() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      w.write("abcdefg".toCharArray(), 1, 5);
      assertEquals("bcdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteString() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abcdefg");
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteString_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abc \u20AC def");
      assertEquals("abc \u20AC def", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteString_encodedBufferOverflow() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      w.write("abcdefg");
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteStringRange() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abcdefg", 1, 5);
      assertEquals("bcdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteStringRange_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abc \u20AC def", 1, 7);
      assertEquals("bc \u20AC de", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteStringRange_all() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.write("abcdefg", 0, 7);
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testWriteStringRange_encodedBufferOverflow() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      w.write("abcdefg", 1, 5);
      assertEquals("bcdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequence() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.append("abcdefg");
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequence_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.append("abc \u20AC def");
      assertEquals("abc \u20AC def", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequence_encodedBufferOverflow() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      w.append("abcdefg");
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequenceRange() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.append("abcdefg", 1, 6);
      assertEquals("bcdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequenceRange_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.append("abc \u20AC def", 1, 8);
      assertEquals("bc \u20AC de", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequenceRange_all() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.append("abcdefg", 0, 7);
      assertEquals("abcdefg", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendCharSequenceRange_encodedBufferOverflow() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(4))) {
      w.append("abcdefg", 1, 6);
      assertEquals("bcdef", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendChar() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.append('a');
      assertEquals("a", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testAppendChar_multiByteEncoding() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      /* Euro sign. */
      w.append('\u20AC');
      assertEquals(3, stream.size());
      assertEquals("\u20AC", stream.toString(UTF_8.name()));
    }
  }

  @Test
  public void testFlush() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.flush();
    }
  }

  @Test
  public void testClose() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.close();
    }
  }

  @Test
  public void testClose_twice() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.close();
      w.close();
    }
  }

  @Test
  public void testClose_methodsThrowIOExceptionAfterClose() throws IOException {
    ByteArrayOutputStream stream = new ByteArrayOutputStream(32);
    try (Writer w = newWriter(stream, UTF_8.newEncoder(), ByteBuffer.allocate(128))) {
      w.close();
      /* @formatter:off */
      try { w.write((int)'a'); fail(); } catch (IOException e) {}
      try { w.write("abc".toCharArray()); fail(); } catch (IOException e) {}
      try { w.write("abc".toCharArray(), 0, 2); fail(); } catch (IOException e) {}
      try { w.write("abc"); fail(); } catch (IOException e) {}
      try { w.write("abc", 0, 2); fail(); } catch (IOException e) {}
      try { w.append("abc"); fail(); } catch (IOException e) {}
      try { w.append("abc", 0, 2); fail(); } catch (IOException e) {}
      try { w.append('a'); fail(); } catch (IOException e) {}
      try { w.flush(); fail(); } catch (IOException e) {}
      /* @formatter:on */
    }
  }

  private static UnbufferedWritableByteChannelWriter newWriter(OutputStream stream,
      CharsetEncoder encoder, ByteBuffer encoded) {
    return new UnbufferedWritableByteChannelWriter(Channels.newChannel(stream), encoder, encoded);
  }
}
